open C

exception NotImplemented
exception UndefinedSemantics

module type DOMAIN =
	sig
		type elt (* the type of abstract domain elements *) 
		val bot: elt
		val join: elt -> elt -> elt (* least upper bound *)
		val leq: elt -> elt -> bool (* less than or equal to *)
		val string_of_elt: elt -> string
		val add: elt -> elt -> elt (* addition between elements *)
		val sub: elt -> elt -> elt (* subtraction between elements *)
		val mul: elt -> elt -> elt (* multiplication between elements *)
	end

module type INTSET_DOMAIN =
	sig
		include DOMAIN
		val filter: (int -> bool) -> elt -> elt
		val make: int list -> elt
	end

module type PARITY_DOMAIN =
	sig
		include DOMAIN
		val top: elt
		val meet: elt -> elt -> elt (* greatest lower bound *)
		val make: int -> elt
	end
	
module type INTERVAL_DOMAIN =
	sig
		include DOMAIN
		type bound = Z of int | Pinfty | Ninfty
		val top: elt
		val meet: elt -> elt -> elt (* greatest lower bound *)
		val make: bound -> bound -> elt
		val widen: elt -> elt -> elt 
		val narrow: elt -> elt -> elt 
	end
	
module IntCPO = 
struct
	type elt = Bot | Int of int
	let bot = Bot
	let join t1 t2 = 
		match t1, t2 with 
		| Bot, _ -> t2 
		| _, Bot -> t1
		| _ -> raise (Failure "can't join two different integers!")
	
	let of_int n = Int n 
	
	let leq t1 t2 =
		match t1, t2 with 
		| Bot, _ -> true 
		| _, Bot -> false
		| (Int n1, Int n2) -> n1 <= n2   

	let string_of_elt t =
		match t with 
		| Bot -> "bot"
		| Int n -> string_of_int n 

	let add t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 + n2) 
	
	let sub t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 - n2)

	let mul t1 t2 =
		match t1, t2 with 
		| Bot, _ -> Bot
		| _, Bot -> Bot
		| Int n1, Int n2 -> Int (n1 * n2)
end

(* helper function for IntSet *)
(* return the cartesian product of alst and blst *)
let rec cartesian alst blst =
  let rec cartesian_sub a blst =
      match blst with [] -> [] | hd :: tl -> (a, hd) :: (cartesian_sub a tl)
  in
  match alst with
  | hd::tl -> (cartesian_sub hd blst) @ (cartesian tl blst)
  | [] -> []
	
module IntsetCPO : INTSET_DOMAIN = 
struct
	(* A data structure for integer sets *)
	(* see https://ocaml.org/api/Set.Make.html *)
	include Set.Make (struct 
		type t = int
		let compare = Pervasives.compare 
	end)
	type elt = t 
	let make = raise NotImplemented 
	let bot = raise NotImplemented
	let join = raise NotImplemented
	let leq = raise NotImplemented
	let add = raise NotImplemented
	let sub = raise NotImplemented
	let mul = raise NotImplemented
	let string_of_elt t = 
		(fold (fun n str -> Printf.sprintf "%s, %d" str n) t "{") ^ "}"  		 
end


module ParityCPO : PARITY_DOMAIN = struct
	type elt = Even | Odd | Top | Bot
	let top = raise NotImplemented
	let bot = raise NotImplemented 
	let make n = raise NotImplemented
	let join t1 t2 = raise NotImplemented
	let meet t1 t2 = raise NotImplemented
	let leq t1 t2 = raise NotImplemented
	let string_of_elt t = 
		match t with 
		| Bot -> "bot_parity"
		| Top -> "top_parity"
		| Even -> "even"
		| Odd -> "odd"	

	let add t1 t2 = raise NotImplemented
	let sub = raise NotImplemented
	let mul t1 t2 = raise NotImplemented
end


module IntervalCPO : INTERVAL_DOMAIN = 
struct
	type bound = Z of int | Pinfty | Ninfty
	type elt = Bot | Intv of (bound * bound)
	let top = raise NotImplemented
	let bot = raise NotImplemented 
	let make lb ub = raise NotImplemented
  let leq t1 t2 = raise NotImplemented 
	let join t1 t2 = raise NotImplemented
	let meet t1 t2 = raise NotImplemented
	let widen t1 t2 = raise NotImplemented
	let narrow t1 t2 = raise NotImplemented
			
	let string_of_bound i = 
		match i with
    | Pinfty -> "oo"
		| Ninfty -> "-oo"
		| Z n -> string_of_int n
		
	let string_of_elt = function
    Bot -> "bot_interval"
  	| Intv (i1, i2) -> "["^(string_of_bound i1)^", "^(string_of_bound i2)^"]"

	let add t1 t2 = raise NotImplemented
	
	let sub t1 t2 = raise NotImplemented
		
	let mul t1 t2 = raise NotImplemented
end


module VarMap = Map.Make (struct 
		type t = var
		let compare = String.compare
	end)
	 
module MakeMemCPO (D : DOMAIN) = 
struct
	(* see https://ocaml.org/api/Map.Make.html *)
	include VarMap
	
	type t = D.elt VarMap.t (* type : var -> D.elt *)

	let val_compare t1 t2 = 
		if (D.leq t1 t2) && (D.leq t2 t1) then 0
		else if (D.leq t1 t2) then -1 
		else 1
			
	let find x t = 
		try 
			find x t 
		with Not_found -> D.bot
	let keys m = fold (fun k _ acc -> k :: acc) m []
	
	(* join two memories *)
  let join m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (D.join v1 v2) m   
  	) empty vars	      
  
	(* widen two memories *)
  let widen widen_op m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (widen_op v1 v2) m   
  	) empty vars	      
	
	(* narrow two memories *)
  let narrow narrow_op m1 m2 = 
  	let vars = (keys m1) @ (keys m2) in
  	List.fold_left (fun m var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		add var (narrow_op v1 v2) m   
  	) empty vars	 
				
  (* check if m1 <= m2 for two memories m1 m2 *)
  let leq m1 m2 = 
  	let vars = (keys m1) in
  	List.fold_left (fun acc var -> 
  		let v1 = find var m1 in  
  		let v2 = find var m2 in
  		acc && (D.leq v1 v2)
  	) true vars	      
	
	let eq m1 m2 = (leq m1 m2) && (leq m2 m1) 	
  
  let string_of_t mem = 
  	fold (fun x v str -> 
			Printf.sprintf "%s\n%s -> %s" str x (D.string_of_elt v)
		) mem ""
end 
	
(* Data structure for various kinds of memory states *)
module Mem = MakeMemCPO(IntCPO) 
module IntsetMem = MakeMemCPO(IntsetCPO)
module ParityMem = MakeMemCPO(ParityCPO)
module IntervalMem = MakeMemCPO(IntervalCPO)
